/* ****************************************************************************
  RC5 IR frames decoding.

  François Pessaux (c) - 2014.
  This piece of code can be freely modified and/or modified.

  Uses 36 Khz signal, i.e. works only with TV telecomands. For other
  telecomands, the bit duration constant has to be changed (and may be the
  tolerance delay.
  *************************************************************************** */

/* Led plugged on digital pin 13. */
#define PIN_LED (13)

#define DEBUG

#ifdef DEBUG
#define PIN_LED_DEBUG1 (11)
#define PIN_LED_DEBUG2 (12)
#endif

/* Output of IR receiver on digital pin 2. Implies to use interrupt #0. */
#define PIN_INPUT (2)
#define IT_NUM (0)

/* In milliseconds. */
#define FRAME_DURATION (25)          /** Duration of a complete frame. */
#define FRAME_REPEAT_INTERV (114)    /** Minimal interval between 2 frames. */

/* In microseconds. */
/* Durations of various bits multiples and sub-multiples. */
#define BIT_DURATION (1778)
#define HALF_BIT_DURATION (BIT_DURATION >> 1)
#define ONE_N_HALF_BIT_DURATION (BIT_DURATION + HALF_BIT_DURATION)
#define TWO_BITS_DURATION (BIT_DURATION << 1)

/* Number of microseconds accepted around the theoretical timing (i.e. +/-). */
#define TIMING_TOLERANCE (30)



/* Attention: Do not put it in the loop () function since this function is
   exactly like a while loop: any local variable is reset at the beginning
   of the function ! Hence, putting it in the loop () function would make
   it reset to 0 at each run of loop () ! */
/** Date of the previous logical rising edge. */
volatile unsigned long prev_change_date = 0 ;

/** Previous decoded bit (if some). Initialised by the automaton when entering
   the DS_Decoding state for the first time. */
volatile unsigned int prev_bit ;

/** Number of bits acquired for the current frame. Initialised by the automaton
   when entering the DS_Decoding state for the first time. */
volatile unsigned int nb_read_bits ;



/** Type representing the 4 possible states of the decoding automaton. */
enum decoder_state_t {
  DS_Idle,
  DS_Decoding,
  DS_Error,
  DS_Done
};



/** Initiale state of the decoding automaton.
   DS_Idle -> DS_Decoding : on first logical rising edge.
   DS_Decoding -> DS_Decoding: consistent bits and < then 14 bits decoded.
   DS_Decoding -> DS_Done: consistent bits and = 14 bits decoded.
   DS_Done -> DS_Idle: external processing ate the bits and request a new
     decoding.
   xxx -> DS_Error: any other cases. */
volatile enum decoder_state_t state = DS_Idle ;



/* 32 bits buffer storign the read bits. Initialized when starting decoding. */
volatile unsigned int read_bits ;



/** Elapsed time between 2 consecutive edges.
    TODO: put it in the ISR ! Is currently there for debug purpose. */
volatile unsigned long change_delta ;



/** Inlined code to disactivate the timer 1 used to detect last 0-bit if some
    by timeout. */
#define STOP_TIMER \
  { TCCR1B = 0 ;                  /* Stop the timer. */                 \
    TIMSK1 &= ~ (1 << OCIE1A) ; } /* Disable the timer interrupt (overkill?). */



/** **************************************************************************
    Initialize the timer to make it firing after 1/2 bit duration. This
    function is launched afte w ereceived the 13th bit, to make sure that if
    the last (14th) bit is a 0 -- in this case we will never receive any
    other logical rising edge -- this bit will be taken into account.
    ************************************************************************* */
void start_timer1_half_bit ()
{
#ifdef DEBUG
  /* Turn on the debug LED monitoring wether the timer is running. */
  digitalWrite (PIN_LED_DEBUG2, HIGH) ;
#endif
  /* Clear current elapsed time. */
  TCNT1  = 0x0 ;
  /* Use Timer/Counter Control Register 1. */
  TCCR1A = 0x0 ;   /* No significant bits to set. */
  TCCR1B =
    (1 << WGM12) |   /* Set CTC mode (Clear Timer on Compare match). */
    (1 << CS12) ;    /* Set 256 prescaler. */

#define MCU_FREQ_HZ (16000000)
#define TIMER_PRESCALE (64)
#define USER_DELAY_S ((HALF_BIT_DURATION + TIMING_TOLERANCE) * 10e-6)
  /* Set the value to compare with in the Output Compare Register.
     We should ensure that the obtained value is not >= 2^(register size).
     The formula is magic directly coming from the ATmega368 datasheet ^_^
     OCR1A = (MCU_FREQ_HZ / TIMER_PRESCALE) / USER_TIMER_FREQ_HZ
     which is transformed to get the init value depending on the delay instead
     of the frequency in
     OCR1A = (MCU_FREQ_HZ / TIMER_PRESCALE) * USER_DELAY_S */
  OCR1A = (MCU_FREQ_HZ / TIMER_PRESCALE) * USER_DELAY_S ;
  /* Enable timer compare interrupt in the Timer/Counter Interrupt Mask
     Register 1. */
  TIMSK1 |= (1 << OCIE1A) ;

}



/** **************************************************************************
    ISR called when the timer 1 fires. This timer is tuned on 1/2 bit duration
    (+ a tolerance) and is initialized when the 13th is read. This mechanism
    is needed to trigger a timeout if the last bit of a frame is a 0. In effect,
    in this case, since the IR interrupt waits for logical rising edges, a 0
    never raise a rising edge, hence the last 0-bit would never arrive.
    ************************************************************************* */
ISR (TIMER1_COMPA_vect)
{
  STOP_TIMER ;
  read_bits = read_bits << 1 ;   /* Shift a 0 on the right of the read bits. */
  nb_read_bits ++ ;              /* One more bit is read. */
  state = DS_Done ;              /* End of decoding state. */
#ifdef DEBUG
  /* Turn off the debug LED monitoring wether the timer is running. */
  digitalWrite (PIN_LED_DEBUG2, LOW) ;
#endif
}



/** **************************************************************************
    Attention: Because IR receiver has electrically "0" on HIGH and "1" on LOW,
    HIGH -> LOW means "logical rising" edge
    and
    LOW -> HIGH means "logical falling" edge.
    Since we are looking for logical rising edges, this ISR is attached to the
    physical falling edge.
    This ISR is in charge of decoding incomming signal to extract its bits.
    ************************************************************************* */
void pin_2_logical_rising_occurred ()
{
  /* Get current date of change. */
  unsigned long change_date = micros () ;

  /* React depending on the current state of the automaton. */
  switch (state) {
  case DS_Idle:
    /* Record the change date. */
    prev_change_date = change_date ;
    /* Initiate decoding. */
    state = DS_Decoding ;
    /* First bit 1 is acquired. */
    prev_bit = 1 ;
    nb_read_bits = 1 ;
    /* First bit is a 1. We will shift bits on the fly. */
    read_bits = 1 ;
    break ;

  case DS_Done:  /* Fall through. Nothing to do since already all bits eaten. */
  case DS_Error: /* Do nothing since already in error. */
    break ;

  case DS_Decoding:
    /* Compute the elapsed time since the previous change. */
    change_delta = change_date - prev_change_date ;
    /* Update the "new old" date of change. */
    prev_change_date = change_date ;

    /* Now switch on the delta value. Three cases are interesting: 1t, 1.5t
       and 2t, t being the buration of a bit. Since timing is never exact, we
       tolerate +/- TIMING_TOLERANCE microseconds. */
    if ((change_delta >= BIT_DURATION - TIMING_TOLERANCE) &&
        (change_delta <= BIT_DURATION + TIMING_TOLERANCE)) {
      /* Delta is 1 t. bit(i+1) = bit(i). Hence prev_bit is left unchanged. */
      read_bits = (read_bits << 1) | prev_bit ;
      nb_read_bits ++ ;
    }

    else if ((change_delta >= ONE_N_HALF_BIT_DURATION - TIMING_TOLERANCE) &&
        (change_delta <= ONE_N_HALF_BIT_DURATION + TIMING_TOLERANCE)) {
      /* Delta is 1.5 t.
         If bit(i) = 1 then bit(i+1) = 0.
         Else (if bit(i) = 0) then bit(i+1) = 0 and bit(i+2) = 1. */
      if (prev_bit) {
        read_bits = read_bits << 1 ;
        prev_bit = 0 ;
        nb_read_bits ++ ;
      }
      else {
        read_bits = (read_bits << 2) | 1 ;
        prev_bit = 1 ;
        nb_read_bits += 2 ;
      }
    }

    else if ((change_delta >= TWO_BITS_DURATION - TIMING_TOLERANCE) &&
        (change_delta <= TWO_BITS_DURATION + TIMING_TOLERANCE)) {
      /* Delta is 2t.
         If bit(i) = 1 then bit(i+1) = 0 and bit(i+2) = 1.
         Else garbage. */
      if (prev_bit) {
        read_bits = read_bits << 2 | 1 ;
        prev_bit = 1 ;
        nb_read_bits += 2 ;
      }
    }

    else {
      /* Else other deltas. Then garbage. */
      STOP_TIMER ;
      state = DS_Error ;
      return ;
    }

    /* If not exited because of the above else, then we are not in error,
       hence we check for the need to launch the timer and for the end of
       decoding reached. */
    if (nb_read_bits == 13) start_timer1_half_bit () ;
    else if (nb_read_bits == 14) {
      STOP_TIMER ;
      state = DS_Done ;
    }
    break ;        /* End of case DS_Decoding. */
  }                /* End of switch. */
}



/** **************************************************************************
    Setup. Configure pins (out for LED and in for IR receiver) and attach the
    ISR to the interrupt.
    ************************************************************************* */
void setup ()
{
  pinMode (PIN_LED, OUTPUT) ;    /* Led on an output. */
#ifdef DEBUG
  pinMode (PIN_LED_DEBUG1, OUTPUT) ;    /* Led for debug on an output. */
  pinMode (PIN_LED_DEBUG2, OUTPUT) ;    /* Led for debug on an output. */
#endif
  pinMode (PIN_INPUT, INPUT) ;   /* IR receiver on an input. */
  attachInterrupt (IT_NUM, pin_2_logical_rising_occurred, FALLING) ;
  Serial.begin (9600) ;
}



/** **************************************************************************
    Mainloop. Mostly jabbering when the decoding process is ended or in error.
    Also resets the de coding automaton after success / error.
    In case of error, we simply wait a frame duration to be sure to land in
    a "no-signal" area and not in a following frame.
    ************************************************************************* */
void loop ()
{
  switch (state) {
  case DS_Idle:
    digitalWrite (PIN_LED, LOW) ;
    break ;

  case DS_Decoding:
    digitalWrite (PIN_LED, HIGH) ;
    break ;

  case DS_Error:
    /* Error case. Print report and wait for a complete frame duration to get
       re-synchronized. */
    digitalWrite (PIN_LED, LOW) ;
#ifdef DEBUG
    digitalWrite (PIN_LED_DEBUG1, HIGH) ;
#endif
    Serial.print ("Error. Nb got bits: ") ;
    Serial.println (nb_read_bits, DEC) ;
    Serial.print ("Delta: ") ;
    Serial.println (change_delta, DEC) ;
    Serial.print ("Prev change date: ") ;
    Serial.println (prev_change_date, DEC) ;
    Serial.print ("Value: ") ;
    Serial.println (read_bits, BIN) ;
    Serial.println ("") ;
    nb_read_bits = 0 ;
    read_bits = 0 ;
    delay (FRAME_DURATION) ;
    /* Reset the decoder to be able to receive another frame. */
#ifdef DEBUG
    digitalWrite (PIN_LED_DEBUG1, LOW) ;
#endif
    state = DS_Idle ;
    break ;

  case DS_Done:
    /* Do whatever you want with the bits of the frame. */
    Serial.print ("Read bits: 0b") ;
    Serial.print (read_bits, BIN) ;
    Serial.print (", 0x") ;
    Serial.println (read_bits, HEX) ;
    /* Reset the decoder to be able to receive another frame. */
    state = DS_Idle ;
    break ;
  }
}
